/***************************************************************************
 *            ovcconnection.c
 *
 *  Tue Jun 20 14:13:49 2006
 *  Copyright  2006  Andrea Del Signore
 *  sejerpz@tin.it
 ***************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "ovc-connection.h"

#include <glib.h>
#include <glib/gi18n.h>
#include <gnet.h>

static void ovc_connection_class_init (OvcConnectionClass * klass);
static void ovc_connection_init (OvcConnection * sp);
static void ovc_connection_finalize (GObject * object);
static void ovc_connection_dispose (GObject * object);

gchar *status_descriptions[OVC_CONNECTION_STATUS_LAST] = 
{ N_("Unknown"), 
  N_("Querying"), 
  N_("Not Connected"), 
  N_("Connecting"), 
  N_("Connected"), 
  N_("Closing"), 
  N_("Initializing") 
};

typedef enum
{
	RESPONSE_WAITING,
	RESPONSE_RECEIVED,
	RESPONSE_TIMEDOUT,
} RESPONSE_STATUS;

struct _OvcConnectionPrivate
{
	/* Place Private Members Here */
	gchar *name;
	gchar *host;
	gint port;
	OVC_CONNECTION_STATES  status;
	GTcpSocket *socket;
	GIOChannel *socket_io;
	gchar *socket_io_expected_response;
	RESPONSE_STATUS socket_io_expected_response_received;
	gboolean disposed;
};

typedef struct _OvcConnectionSignal OvcConnectionSignal;
typedef enum _OvcConnectionSignalType OvcConnectionSignalType;

enum _OvcConnectionSignalType
{
	/* Place Signal Types Here */
	SIGNAL_TYPE_DATA_RECEIVED,
	SIGNAL_TYPE_CONNECTION_STATUS_CHANGED,
	SIGNAL_TYPE_CONNECTION_STATUS_CHANGING,
	LAST_SIGNAL
};

struct _OvcConnectionSignal
{
	OvcConnection *object;
};

static guint ovc_connection_signals[LAST_SIGNAL] = { 0 };
static GObjectClass *parent_class = NULL;

static void
set_connection_status(OvcConnection *self, OVC_CONNECTION_STATES status)
{
	if (self->priv->status != status)
	{
		g_signal_emit (self, ovc_connection_signals[SIGNAL_TYPE_CONNECTION_STATUS_CHANGING], 0 , status);
		self->priv->status = status;
		g_signal_emit (self, ovc_connection_signals[SIGNAL_TYPE_CONNECTION_STATUS_CHANGED], 0 , status);
	}
}


GType
ovc_connection_get_type ()
{
	static GType type = 0;

	if (type == 0)
	{
		static const GTypeInfo our_info = {
			sizeof (OvcConnectionClass),
			NULL,
			NULL,
			(GClassInitFunc) ovc_connection_class_init,
			NULL,
			NULL,
			sizeof (OvcConnection),
			0,
			(GInstanceInitFunc) ovc_connection_init,
		};

		type = g_type_register_static (G_TYPE_OBJECT,
									   "OvcConnection", &our_info, 0);
	}

	return type;
}

static void
ovc_connection_set_property (GObject * object,
							 guint property_id,
							 const GValue * value, GParamSpec * pspec)
{
	OvcConnection *self = (OvcConnection *) object;

	
	switch (property_id)
	{
	case OVC_CONNECTION_CONSTRUCT_NAME:
		g_free (self->priv->name);
		self->priv->name = g_value_dup_string (value);
		break;
	case OVC_CONNECTION_HOST:
		g_free (self->priv->host);
		self->priv->host = g_value_dup_string (value);
		break;
	case OVC_CONNECTION_PORT:
		self->priv->port = g_value_get_int (value);
		break;
	default:
		/* We don't have any other property... */
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}

static void
ovc_connection_get_property (GObject * object,
							 guint property_id,
							 GValue * value, GParamSpec * pspec)
{
	OvcConnection *self = (OvcConnection *) object;

	switch (property_id)
	{
		case OVC_CONNECTION_CONSTRUCT_NAME:
			g_value_set_string (value, self->priv->name);
			break;
		case OVC_CONNECTION_STATUS:
			g_value_set_int (value, self->priv->status);
			break;		
		case OVC_CONNECTION_HOST:
			g_value_set_string (value, self->priv->host);
			break;
		case OVC_CONNECTION_PORT:
			g_value_set_int (value, self->priv->port);
			break;
		default:
			/* We don't have any other property... */
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
			break;
	}
}

static void
ovc_connection_class_init (OvcConnectionClass * klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	GParamSpec *connection_param_spec;
	GType param_types[1];
	
	parent_class = g_type_class_peek_parent (klass);
	object_class->finalize = ovc_connection_finalize;
	object_class->finalize = ovc_connection_dispose;
	
	param_types[0] = G_TYPE_STRING;
	
	ovc_connection_signals[SIGNAL_TYPE_DATA_RECEIVED] =   g_signal_newv("data-received",
                 G_TYPE_FROM_CLASS (object_class),
                 G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE | G_SIGNAL_NO_HOOKS,
                 NULL /* class closure */,
                 NULL /* accumulator */,
                 NULL /* accu_data */,
                 g_cclosure_marshal_VOID__STRING,
                 G_TYPE_NONE /* return_type */,
                 1     /* n_params */,
                 param_types  /* param_types */); 

	param_types[0] = G_TYPE_INT;
	ovc_connection_signals[SIGNAL_TYPE_CONNECTION_STATUS_CHANGED] =   g_signal_newv("status-changed",
                 G_TYPE_FROM_CLASS (object_class),
                 G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE | G_SIGNAL_NO_HOOKS,
                 NULL /* class closure */,
                 NULL /* accumulator */,
                 NULL /* accu_data */,
                 g_cclosure_marshal_VOID__ENUM,
                 G_TYPE_NONE /* return_type */,
                 1     /* n_params */,
                 param_types  /* param_types */); 
	ovc_connection_signals[SIGNAL_TYPE_CONNECTION_STATUS_CHANGING] =   g_signal_newv("status-changing",
                 G_TYPE_FROM_CLASS (object_class),
                 G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE | G_SIGNAL_NO_HOOKS,
                 NULL /* class closure */,
                 NULL /* accumulator */,
                 NULL /* accu_data */,
                 g_cclosure_marshal_VOID__ENUM,
                 G_TYPE_NONE /* return_type */,
                 1     /* n_params */,
                 param_types  /* param_types */); 

	
	object_class->set_property = ovc_connection_set_property;
	object_class->get_property = ovc_connection_get_property;
	
	connection_param_spec = g_param_spec_string ("name",
												 "Connection name construct prop",
												 "Set Ovc connection's name",
												 "no-name-set"
												 /* default value */ ,
												 G_PARAM_CONSTRUCT |
												 G_PARAM_READWRITE);
	
	g_object_class_install_property (object_class,
									 OVC_CONNECTION_CONSTRUCT_NAME,
									 connection_param_spec);

	connection_param_spec = g_param_spec_string ("host",
							 "hostname",
							 "Set Ovc connection's host",
							 "no-name-set"
							 /* default value */ ,
							 G_PARAM_READWRITE);
	
	g_object_class_install_property (object_class,
									 OVC_CONNECTION_HOST,
									 connection_param_spec);

 	connection_param_spec = g_param_spec_int ("port",
							 "port number",
							 "Set Ovc connection's port",
							 0,65535,0,
							 G_PARAM_READWRITE);

 	g_object_class_install_property (object_class,
									 OVC_CONNECTION_PORT,
									 connection_param_spec);

	connection_param_spec = g_param_spec_int ("status",
												 "Connection status prop",
												 "Get Ovc connection status",
												 0,OVC_CONNECTION_STATUS_CONNECTED,OVC_CONNECTION_STATUS_UNKNOWN,
												 G_PARAM_READABLE);
	
	g_object_class_install_property (object_class,
									 OVC_CONNECTION_STATUS,
									 connection_param_spec);
	
}

static void
ovc_connection_init (OvcConnection * obj)
{
	obj->priv = g_new0 (OvcConnectionPrivate, 1);
	/* Initialize private members, etc. */
	obj->priv->disposed = FALSE;
	obj->priv->status = OVC_CONNECTION_STATUS_UNKNOWN;
	obj->priv->host = NULL;
	obj->priv->port = 0;
	obj->priv->socket = NULL;
	obj->priv->socket_io_expected_response = NULL;
}

static void
ovc_connection_dispose(GObject *object)
{
	OvcConnection *cobj;
	cobj = OVC_CONNECTION (object);
	
	if (cobj->priv->disposed)
		return;
	
	cobj->priv->disposed = TRUE;
}

static void
ovc_connection_finalize (GObject * object)
{
	OvcConnection *cobj;
	cobj = OVC_CONNECTION (object);

	/* Free private members, etc. */
	g_free (cobj->priv->name);
	g_free (cobj->priv->socket_io_expected_response);
	g_free (cobj->priv);
	
	G_OBJECT_CLASS (parent_class)->finalize (object);
	
}




static gboolean
async_client_iofunc (GIOChannel * iochannel, GIOCondition condition,
						 gpointer data)
{
	OvcConnection *ovcc = (OvcConnection *) data;
	GTcpSocket *async_socket = ovcc->priv->socket;
	/* Check for socket error */
	if (condition & (G_IO_ERR | G_IO_HUP | G_IO_NVAL))
	{
		return FALSE;
	}

	/* Check for data to be read */
	if (condition & G_IO_IN)
	{
		GIOError error;
		gchar buffer[1024];
		guint bytes_read;

		/* Read the data into our buffer */
		error = g_io_channel_read (iochannel, buffer,
								   sizeof (buffer), &bytes_read);

		/* Check for stdin error */
		if (error != G_IO_ERROR_NONE)
		{
			return FALSE;
		}

		/* Check for EOF */
		else if (bytes_read == 0)
		{
			gnet_tcp_socket_delete (async_socket);
			return FALSE;
		}

		/* Otherwise, print */
		else
		{
			buffer[bytes_read] = 0;
			g_signal_emit (ovcc, ovc_connection_signals[0], 0 , buffer);
			if (g_pattern_match_simple("*STATE:*,CONNECTED,SUCCESS,*", buffer))
				set_connection_status(ovcc, OVC_CONNECTION_STATUS_CONNECTED);
			else if (g_pattern_match_simple("*STATE:*,RECONNECTING,*error*", buffer))
				set_connection_status(ovcc, OVC_CONNECTION_STATUS_NOT_CONNECTED);
			else if (g_pattern_match_simple("*STATE:*,RECONNECTING,SIGHUP*", buffer))
				set_connection_status(ovcc, OVC_CONNECTION_STATUS_NOT_CONNECTED);
			else if (g_pattern_match_simple("*STATE:*,WAIT", buffer))
				set_connection_status(ovcc, OVC_CONNECTION_STATUS_CONNECTING);
			else if (ovcc->priv->socket_io_expected_response != NULL &&
					 g_pattern_match_simple(ovcc->priv->socket_io_expected_response, buffer))
			{
				ovcc->priv->socket_io_expected_response_received = RESPONSE_RECEIVED;
			}
		}
		
	}

	return TRUE;

}



static gboolean
ovc_connection_write_timeout(gpointer data)
{
	OvcConnection *self = (OvcConnection *) data;
	if (self->priv->socket_io_expected_response_received == RESPONSE_WAITING)
		self->priv->socket_io_expected_response_received = RESPONSE_TIMEDOUT;
	
	return FALSE;
}

static gboolean
ovc_connection_write_and_wait_response(OvcConnection *self, gchar *data, const gchar *response, gint timeout)
{
	guint bytes_written = 0;
	gchar *buffer = g_strdup_printf("%s\n",data);
	
	self->priv->socket_io_expected_response = g_strdup(response);
	self->priv->socket_io_expected_response_received = RESPONSE_WAITING;
	gnet_io_channel_writen (self->priv->socket_io, buffer, g_utf8_strlen(buffer,-1), &bytes_written);
	guint timer_id = g_timeout_add(timeout, ovc_connection_write_timeout, (gpointer) self);
	
	while(self->priv->socket_io_expected_response_received == RESPONSE_WAITING)
		 g_main_context_iteration(NULL, FALSE);
	
	g_source_remove(timer_id);

	g_free(self->priv->socket_io_expected_response);
	self->priv->socket_io_expected_response = NULL;
	g_free(buffer);
	return (self->priv->socket_io_expected_response_received != RESPONSE_TIMEDOUT ? TRUE : FALSE);
}


static void
async_client_connfunc (GTcpSocket * socket,
					   GTcpSocketConnectAsyncStatus status, gpointer data)
{
	OvcConnection *ovcc = (OvcConnection *) data;

	ovcc->priv->socket = socket;
	
	if (status != GTCP_SOCKET_CONNECT_ASYNC_STATUS_OK)
	{
		set_connection_status(ovcc, OVC_CONNECTION_STATUS_UNKNOWN);
		return;
	}
	gchar *name;
	g_object_get (ovcc, "name", &name, NULL);
	g_free (name);
	/* Socket stream */
	ovcc->priv->socket_io = gnet_tcp_socket_get_io_channel (socket);
	g_io_add_watch (ovcc->priv->socket_io, G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
					async_client_iofunc, ovcc);
	
	set_connection_status(ovcc, OVC_CONNECTION_STATUS_INITIALIZING);
}

void
ovc_connection_control_channel_open (OvcConnection *self)
{
	set_connection_status(self, OVC_CONNECTION_STATUS_QUERYING);
	gnet_tcp_socket_connect_async (self->priv->host, self->priv->port, async_client_connfunc,
								   self);
	
	while(self->priv->status == OVC_CONNECTION_STATUS_QUERYING)
	    g_main_context_iteration(NULL, FALSE);
	
	if (self->priv->status != OVC_CONNECTION_STATUS_UNKNOWN)
	{
		/* Check status */
		if (ovc_connection_write_and_wait_response(self,"state","*CONNECTING*",5000))
		{
			set_connection_status(self, OVC_CONNECTION_STATUS_NOT_CONNECTED);			
		}
		else if (ovc_connection_write_and_wait_response(self,"state","*CONNECTED*",5000))
		{
			set_connection_status(self, OVC_CONNECTION_STATUS_CONNECTED);			
		}
		else
			set_connection_status(self, OVC_CONNECTION_STATUS_UNKNOWN);

		if (self->priv->status != OVC_CONNECTION_STATUS_UNKNOWN)
		{
			ovc_connection_write_and_wait_response(self, "state on", "SUCCESS*",5000);
			ovc_connection_write_and_wait_response(self, "log off", "SUCCESS*",5000);
		}
	}
}

OvcConnection *
ovc_connection_new (const gchar * name)
{
	OvcConnection *obj;

	obj = OVC_CONNECTION (g_object_new (OVC_TYPE_CONNECTION, "name", name, NULL));
	return obj;
}

gboolean
ovc_connection_write(OvcConnection *self, gchar *data)
{
	guint bytes_written = 0;
	gchar *buffer = g_strdup_printf("%s\n",data);
	guint length = g_utf8_strlen(buffer,-1);
	
	gnet_io_channel_writen (self->priv->socket_io, buffer, length, &bytes_written);
	g_free(buffer);
	return (bytes_written == length ? TRUE : FALSE);
}

void
ovc_connection_open (OvcConnection *self)
{
	g_assert(self->priv->status == OVC_CONNECTION_STATUS_NOT_CONNECTED);
	set_connection_status(self, OVC_CONNECTION_STATUS_CONNECTING);
	
	ovc_connection_write(self, "hold release");
}

void
ovc_connection_close(OvcConnection *self)
{
	g_assert(self->priv->status == OVC_CONNECTION_STATUS_CONNECTED);
	
	gchar *data = "signal SIGHUP";
	set_connection_status(self, OVC_CONNECTION_STATUS_CLOSING);
	ovc_connection_write(self, data);
}

