/***************************************************************************
 *            ovc-connections.c
 *
 *  Thu Jun 22 21:13:56 2006
 *  Copyright  2006  Andrea Del Signore
 *  sejerpz@tin.it
 ***************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "ovc-connections.h"
#include <string.h>
#include <stdlib.h>
#include <errno.h>

static void ovc_connections_class_init(OvcConnectionsClass *klass);
static void ovc_connections_init(OvcConnections *sp);
static void ovc_connections_finalize(GObject *object);

static void 
xmlconfig_start_element_handler(GMarkupParseContext *context,
								  const gchar         *element_name,
								  const gchar        **attribute_names,
								  const gchar        **attribute_values,
								  gpointer             user_data,
								  GError             **error);

static void 	
xmlconfig_end_element_handler(GMarkupParseContext *context,
							  const gchar         *element_name,
							  gpointer             user_data,
							  GError             **error);

static void 
xmlconfig_text_handler(GMarkupParseContext *context,
						const gchar         *text,
						gsize                text_len,  
						gpointer             user_data,
						GError             **error);

struct _OvcConnectionsPrivate {
	/* Place Private Members Here */
	GList *items;
	gboolean disposed;
};

typedef struct _OvcConnectionsSignal OvcConnectionsSignal;
typedef enum _OvcConnectionsSignalType OvcConnectionsSignalType;

enum _OvcConnectionsSignalType {
	/* Place Signal Types Here */
	SIGNAL_TYPE_ITEM_ADDED,
	SIGNAL_TYPE_ITEM_REMOVED,
	SIGNAL_TYPE_LIST_CLEAR,
	LAST_SIGNAL
};

enum
{
	OVC_CONNECTION_ITEMS = 1
};

struct _OvcConnectionsSignal {
	OvcConnections *object;
};

static GMarkupParser xmlconfig_parser =
{
	xmlconfig_start_element_handler,
	xmlconfig_end_element_handler,
	xmlconfig_text_handler,
	NULL,
	NULL
};

static guint ovc_connections_signals[LAST_SIGNAL] = { 0 };
static GObjectClass *parent_class = NULL;

static void
free_connection(gpointer data, gpointer user_data)
{
	gchar *name=NULL;
	
	g_object_get(OVC_CONNECTION(data), "name", &name, NULL);
	g_free(name);
	g_object_unref(OVC_CONNECTION(data));
}

GType
ovc_connections_get_type()
{
	static GType type = 0;

	if(type == 0) {
		static const GTypeInfo our_info = {
			sizeof (OvcConnectionsClass),
			NULL,
			NULL,
			(GClassInitFunc)ovc_connections_class_init,
			NULL,
			NULL,
			sizeof (OvcConnections),
			0,
			(GInstanceInitFunc)ovc_connections_init,
		};

		type = g_type_register_static(G_TYPE_OBJECT, 
			"OvcConnections", &our_info, 0);
	}

	return type;
}

static void
ovc_connections_set_property (GObject * object,
							 guint property_id,
							 const GValue * value, GParamSpec * pspec)
{
//	OvcConnections *self = (OvcConnections *) object;

	switch (property_id)
	{
		default:
			/* We don't have any other property... */
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}

static void
ovc_connections_get_property (GObject * object,
							 guint property_id,
							 GValue * value, GParamSpec * pspec)
{
	OvcConnections *self = (OvcConnections *) object;

	switch (property_id)
	{
		case OVC_CONNECTION_ITEMS:
			g_value_set_pointer (value, self->priv->items);
		break;
		default:
			/* We don't have any other property... */
			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}

static void connection_items_clear(OvcConnections *cobj)
{
	/* Free private members, etc. */
	if (cobj->priv->items != NULL)
	{
		g_list_foreach(cobj->priv->items,free_connection,NULL);
		g_list_free(cobj->priv->items);
		cobj->priv->items = NULL;
		g_signal_emit(OVC_CONNECTIONS(cobj), 
					  ovc_connections_signals[SIGNAL_TYPE_LIST_CLEAR],
					  0);
					  
	}

}

static void
ovc_connections_dispose (GObject *obj)
{
	OvcConnections *cobj;
	cobj = OVC_CONNECTIONS(obj);	

	if (cobj->priv->disposed) 
	{
		/* If dispose did already run, return. */
		return;
	}
	/* Make sure dispose does not run twice. */
	cobj->priv->disposed = TRUE;

	connection_items_clear(cobj);

	/* Chain up to the parent class */
	G_OBJECT_CLASS (parent_class)->dispose (obj);
}


static void
ovc_connections_class_init(OvcConnectionsClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS(klass);
	GParamSpec *connection_param_spec;
	GType param_types[1];
	
	parent_class = g_type_class_peek_parent(klass);
	object_class->dispose = ovc_connections_dispose;
	object_class->finalize = ovc_connections_finalize;
	
	/* Create signals here:	*/
 	
	param_types[0] = G_TYPE_POINTER;
	ovc_connections_signals[SIGNAL_TYPE_ITEM_ADDED] =   g_signal_newv("item-added",
                 G_TYPE_FROM_CLASS (object_class),
                 G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE | G_SIGNAL_NO_HOOKS,
                 NULL /* class closure */,
                 NULL /* accumulator */,
                 NULL /* accu_data */,
                 g_cclosure_marshal_VOID__POINTER,
                 G_TYPE_NONE /* return_type */,
                 1     /* n_params */,
	                   /* param_types */
                 param_types); 
	
	param_types[0] = G_TYPE_POINTER;
	ovc_connections_signals[SIGNAL_TYPE_ITEM_REMOVED] =   g_signal_newv("item-removed",
                 G_TYPE_FROM_CLASS (object_class),
                 G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE | G_SIGNAL_NO_HOOKS,
                 NULL /* class closure */,
                 NULL /* accumulator */,
                 NULL /* accu_data */,
                 g_cclosure_marshal_VOID__POINTER,
                 G_TYPE_NONE /* return_type */,
                 1     /* n_params */,
	                   /* param_types */
                 param_types);
	
	param_types[0] = G_TYPE_NONE;
	ovc_connections_signals[SIGNAL_TYPE_LIST_CLEAR] =   g_signal_newv("list-clear",
                 G_TYPE_FROM_CLASS (object_class),
                 G_SIGNAL_RUN_LAST | G_SIGNAL_NO_RECURSE | G_SIGNAL_NO_HOOKS,
                 NULL /* class closure */,
                 NULL /* accumulator */,
                 NULL /* accu_data */,
                 g_cclosure_marshal_VOID__VOID,
                 G_TYPE_NONE /* return_type */,
                 0     /* n_params */,
	                   /* param_types */
                 param_types);
	
	/* Installing properties handlers */
	object_class->set_property = ovc_connections_set_property;
	object_class->get_property = ovc_connections_get_property;
	
	connection_param_spec = g_param_spec_pointer ("items",
												 "Connection items",
												 "Connection items",
												 G_PARAM_READABLE);
	
	g_object_class_install_property (object_class,
									 OVC_CONNECTION_ITEMS,
									 connection_param_spec);
}

static void
ovc_connections_init(OvcConnections *obj)
{
	obj->priv = g_new0(OvcConnectionsPrivate, 1);
	
	/* Initialize private members, etc. */
	obj->priv->items = NULL;
	obj->priv->disposed = FALSE;
}



static void
ovc_connections_finalize(GObject *object)
{
	OvcConnections *cobj;
	cobj = OVC_CONNECTIONS(object);	
	
	g_free(cobj->priv);
	G_OBJECT_CLASS(parent_class)->finalize(object);
}

OvcConnections *
ovc_connections_new()
{
	OvcConnections *obj;
	
	obj = OVC_CONNECTIONS(g_object_new(OVC_TYPE_CONNECTIONS, NULL));
	
	return obj;
}

gboolean
ovc_connections_load(OvcConnections *self, const gchar *filename)
{
	gchar* content;
	gsize len;
	gboolean result;
	GMarkupParseContext *context;
	
	if (!g_file_test(filename, G_FILE_TEST_EXISTS))
		return TRUE; //no file
	
	if (!(g_file_test(filename, G_FILE_TEST_IS_REGULAR) || 
		  g_file_test(filename, G_FILE_TEST_IS_SYMLINK)))
		return FALSE; //file exists but is not a regular one nor a symlink?!?!
	
	if (!g_file_get_contents (filename, &content, &len, NULL))
	{
		return FALSE;
	}
	
	context = g_markup_parse_context_new(&xmlconfig_parser, 0,
										 self, NULL);

	if ((result = g_markup_parse_context_parse(context, content, len, NULL)))
		result = g_markup_parse_context_end_parse(context, NULL);

	if (!result)
		g_warning("Error parsing the file: %s", filename);
	
	g_markup_parse_context_free(context);
	g_free(content);
	return result;
}

static void 
xmlconfig_start_element_handler(GMarkupParseContext *context,
								  const gchar         *element_name,
								  const gchar        **attribute_names,
								  const gchar        **attribute_values,
								  gpointer             user_data,
								  GError             **error)
{
	OvcConnections *self = (OvcConnections *) user_data;
	
	gchar *folded_el1 = g_utf8_casefold(element_name,-1);
	gchar *folded_el2 = g_utf8_casefold("connection",-1);
	
	if (strcmp(folded_el1,folded_el2) == 0)
	{
		gchar *name = NULL, *host = NULL, *port_text = NULL;
		int i;
		for (i = 0; attribute_names[i] != NULL; i++) 
		{
			if (g_strncasecmp("name",attribute_names[i],4) == 0)
			{
				name = g_strdup(attribute_values[i]);
			} 
			else if (g_strncasecmp("host",attribute_names[i],4) == 0)
			{
				host = g_strdup(attribute_values[i]);
			} 
			else if (g_strncasecmp("port",attribute_names[i],4) == 0)
			{
				port_text = g_strdup(attribute_values[i]);
			}
		}
		if (name != NULL)
		{
			OvcConnection *conn = ovc_connection_new(name);
			errno = 0;
			gint port = strtol(port_text, NULL, 0);
			if (errno)
				port = 0;
			g_object_set(G_OBJECT(conn), "host", host, "port", port, NULL);
			ovc_connections_add(self, conn);
			g_object_unref(G_OBJECT(conn));
			
		}
		g_free(name);
		g_free(host);
		g_free(port_text);
	}
	
	g_free(folded_el1);
	g_free(folded_el2);
}

static void 	
xmlconfig_end_element_handler(GMarkupParseContext *context,
							  const gchar         *element_name,
							  gpointer             user_data,
							  GError             **error)
{
	
}

static void 
xmlconfig_text_handler(GMarkupParseContext *context,
						const gchar         *text,
						gsize                text_len,  
						gpointer             user_data,
						GError             **error)
{
	
}

	

gboolean
ovc_connections_save(OvcConnections *self, const gchar *filename)
{
	gchar *content, *tmp;
	GList *item;
	
	content = g_strdup("<?xml version='1.0'?>\n<connections>\n");

	for(item=self->priv->items; item!=NULL; item = item->next)
	{
		OvcConnection *conn = (OvcConnection *) item->data;
		
		g_object_get(conn,"name",&tmp,NULL);
		gchar *connection = g_markup_printf_escaped ("\t<connection "
													 "name=\"%s\" "
													 "host=\"%s\" "
													 "port=\"%s\""
													 "></connection>\n",
													 tmp,
													 "localhost","7505");
		g_free(tmp);
		
		tmp = g_strconcat(content, connection, NULL);
		g_free(content);
		content = tmp;

		g_free(connection);
	}
	tmp = g_strconcat(content, "</connections>\n", NULL);
	g_free(content);
	content = tmp;
	gboolean res = g_file_set_contents(filename, content, -1, NULL);
	g_free(content);
	return res;
}

OvcConnection*
ovc_connections_add(OvcConnections *self, OvcConnection *connection)
{
	gchar *name;
	g_object_get(G_OBJECT(connection), "name", &name,NULL);
	g_free(name);
	
	self->priv->items = g_list_append(self->priv->items, g_object_ref(G_OBJECT(connection)));
	g_signal_emit(OVC_CONNECTIONS(self), 
				  ovc_connections_signals[SIGNAL_TYPE_ITEM_ADDED],
				  0,
				  G_OBJECT(connection));
	return connection;
}

void
ovc_connections_remove(OvcConnections *self, const OvcConnection *connection)
{
	g_signal_emit(OVC_CONNECTIONS(self), 
				  ovc_connections_signals[SIGNAL_TYPE_ITEM_REMOVED],
				  0,
				  G_OBJECT(connection));
	return;
}
